/**EurothermDrv.c
 * This file list the driver routines used to communicate with
 * a Eurotherm device using the ModBus over serial interface
 * This package depends on the asyn and modbus EPICS packages
 * to communicate with the device
 *
 * Copyright (c) 2009 Stanford University as operator of 
 *     SLAC National Accelerator Laboratory
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <epicsStdio.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <asynDriver.h>
#include <asynOctetSyncIO.h>
#include <asynCommonSyncIO.h>
#include <asynStandardInterfaces.h>
#include <drvModbusAsyn.h>
#include <modbusInterpose.h>
#include <errlog.h>

#define EUROTHERM_MAX_TIMEOUT_MS	1000
#define EUROTHERM_POLL_MS		1000
#define DRIVER_NAME			(__FILE__)
#define LOG_ERROR(str...)		do { \
						errlogPrintf("%s::%s(@%d):", DRIVER_NAME, __func__, __LINE__); \
						errlogPrintf(str); \
						errlogPrintf("\n"); \
					} while(0)
#define MODBUS_READW			3
#define MODBUS_WRITEW			6

#define EUROTHERM_ADDR_TEMP		1
#define EUROTHERM_ADDR_SETPOINT		2
#define EUROTHERM_ADDR_WORKING		5
#define EUROTHERM_ADDR_OUTPUT		4
#define EUROTHERM_ADDR_AUXILIARY	4948
#define EUROTHERM_ADDR_MODE		273
#define EUROTHERM_ADDR_BAR_HI		827
#define EUROTHERM_ADDR_BAR_LO		826
#define EUROTHERM_ADDR_P_BAND		351
#define EUROTHERM_ADDR_I_TIME		352
#define EUROTHERM_ADDR_D_TIME		353


int mkname(char *str, int max, char *asynPortName, int modbusAddress, int bWrite, char *szVariable)
{
	int ret;
	
	if(str == NULL)
		return strlen("Eurotherm__XX_XX_") + strlen(asynPortName) + strlen(szVariable);
	ret = sprintf(str, "Eurotherm_%s_%d_%s_%s", asynPortName, modbusAddress, bWrite? "Wr" : "Rd", szVariable);
	if(ret >= 0) ret++;	/* Add the terminating char */
	return ret;
}

int drvEurothermInit(char *asynPortName, int modbusAddress)
{
	int ret;
	asynStatus asynRet;
	char *str;
	int strmaxlen;

	if (asynPortName == NULL) {
		LOG_ERROR("Invalid asynPortName argument (NULL)");
		return -EINVAL;
	}
	if ((modbusAddress < 0) || (modbusAddress >= 100)){
		LOG_ERROR("Invalid modbusAddress value %d (0<=v<100)", modbusAddress);
		return -EINVAL;
	}
	ret = modbusInterposeConfig(asynPortName, modbusAddress, modbusLinkRTU, EUROTHERM_MAX_TIMEOUT_MS);
	if (ret < 0) {
		LOG_ERROR("modbusInterposeConfig failed with error %d", ret);
		return ret;
	}
	strmaxlen = mkname(NULL, 0, asynPortName, modbusAddress, 0, "MaxNameSize");
	if (strmaxlen < 0) {
		LOG_ERROR("Could not generate modbus names: %s", strerror(errno));
		return -errno;
	}
	str = malloc(strmaxlen);
	if(str == NULL) {
		LOG_ERROR("Could not allocate %d bytes of memory", strmaxlen);
		return -ENOMEM;
	}
	/* Create a modbus read channel for the Temperature */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "Temperature");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_TEMP, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus write channel for the Setpoint */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 1, "SetPoint");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_WRITEW, EUROTHERM_ADDR_SETPOINT, 1,
				0, 1, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the Working point */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "Working");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_WORKING, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the Auxiliary temperature */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "Auxiliary");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_AUXILIARY, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the Output Level */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "OutputLevel");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_OUTPUT, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the Bar HIgh */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "BarHigh");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_BAR_HI, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the Bar Low */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "BarLow");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_BAR_LO, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the AutoManMode */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "AutoManMode");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_MODE, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the P-Band Loop Paramter */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "PBand");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_P_BAND, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the I-Time-Band Loop Paramter */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "ITime");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_I_TIME, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	/* Create a modbus read channel for the D-Time-Band Loop Paramter */
	mkname(str, strmaxlen, asynPortName, modbusAddress, 0, "DTime");
	asynRet = drvModbusAsynConfigure(str, asynPortName,
				MODBUS_READW, EUROTHERM_ADDR_D_TIME, 1,
				0, EUROTHERM_POLL_MS, "");
	if (asynRet != asynSuccess) {
		LOG_ERROR("drvModbusAsynConfigure failed with error %d", asynRet);
		free(str);
		return -EIO;
	}
	free(str);
	return 0;
}

/**EPICS registration API
 * We provide this registrar to make it easy to add a Eurotherm device without
 * having to mess up st.cmd
 */

static const iocshArg ConfigureArg0 = {"Asyn Device Port Name", iocshArgString};
static const iocshArg ConfigureArg1 = {"Modbus Slave Address", iocshArgInt};

static const iocshArg * const drvEurothermInitArgs[2] = {
	&ConfigureArg0,
	&ConfigureArg1,
};

static const iocshFuncDef drvModbusAsynConfigureFuncDef=
                                                    {"drvEurothermInit", 2,
                                                     drvEurothermInitArgs};

static void drvEurothermInitCall(const iocshArgBuf *args)
{
	int ret;

	ret = drvEurothermInit(args[0].sval, args[1].ival);
	if(ret != 0)
		printf("Eurotherm driver init failed with code %d.\n", ret);
}

static void drvEurothermRegister() {
	iocshRegister(&drvModbusAsynConfigureFuncDef  , drvEurothermInitCall );
}

epicsExportRegistrar(drvEurothermRegister);
